extern crate feed_rs;
extern crate maud;
extern crate reqwest;

use feed_rs::model::Entry;
use feed_rs::parser;
use maud::{html, Markup};
use reqwest::blocking::get;
use scraper::{Html, Selector};
use std::cmp::Reverse;
use std::error::Error;
use std::fs;
use std::fs::write;
use std::fs::DirBuilder;
use std::path::Path;

fn fetch_feed(url: &str) -> Result<Vec<Entry>, Box<dyn Error>> {
    let content = get(url)?.text()?;
    let feed = parser::parse(content.as_bytes())?;
    println!("Feed {} returned {} items", url, feed.entries.len());
    Ok(feed.entries)
}

fn create_html_card(entry: &Entry) -> Markup {
    let title = entry
        .title
        .as_ref()
        .map_or_else(|| "".to_string(), |t| t.content.clone());

    let link = entry.links.first().unwrap();

    // Attempt to find the first thumbnail URL in the media content
    let image_src = entry
        .media
        .iter()
        .flat_map(|media| &media.thumbnails)
        .map(|thumbnail| &thumbnail.image.uri)
        .next()
        .map_or_else(|| "".to_string(), |uri| uri.to_string());

    let description = entry.content.as_ref().map_or_else(
        || {
            entry
                .summary
                .as_ref()
                .map_or_else(|| "".to_string(), |summary| summary.content.clone())
        },
        |content| {
            content
                .body
                .as_ref()
                .map_or_else(|| "".to_string(), |body| body.clone())
        },
    );

    let cleaned_description = strip_html_tags(&description);
    let truncated_description = truncate_description(&cleaned_description, 500); // Truncate description to 100 characters

    html! {
        a.card-link href=(link.href) target=("_blank") {
            div.card {
                h2 { (title) }
                @if !image_src.is_empty() {
                    img src=(image_src) alt="Entry image";
                }
                p { (truncated_description) }
            }
        }
    }
}

fn truncate_description(description: &str, max_length: usize) -> String {
    let description_trimmed = description.trim();
    if description_trimmed.len() > max_length {
        format!("{}...", &description_trimmed[..max_length])
    } else {
        description_trimmed.to_string()
    }
}

fn strip_html_tags(html: &str) -> String {
    let document = Html::parse_document(html);
    // Use the wildcard selector to select all nodes and extract their text.
    let selector = Selector::parse("*").unwrap();
    let mut text_content = String::new();

    for element in document.select(&selector) {
        let text = element.text().collect::<Vec<_>>().join(" ");
        text_content.push_str(&text);
        text_content.push(' ');
    }

    text_content.trim().to_string()
}

fn generate_html(entries: Vec<Entry>) -> Markup {
    html! {
        (maud::DOCTYPE)
        html {
            head {
                title { "Anson's Zine" }
                link rel="stylesheet" href="https://cdn.jsdelivr.net/npm/@picocss/pico@1/css/pico.min.css";
                link rel="stylesheet" href="style.css";
            }
            body {
                h1 { "Anson's Aggregated Feed" }
                div class="cards-container" {
                    @for entry in entries {
                        {(create_html_card(&entry))}
                    }
                }
            }
        }
    }
}

fn main() -> Result<(), Box<dyn Error>> {
    let binding = fs::read_to_string("feeds.txt").unwrap();
    let feed_urls: Vec<&str> = binding.lines().collect();

    let mut entries: Vec<Entry> = Vec::new();

    for url in feed_urls {
        match fetch_feed(url) {
            Ok(mut feed_entries) => entries.append(&mut feed_entries),
            Err(e) => println!("Failed to fetch or parse feed {}: {}", url, e),
        }
    }

    // Remove any entries that don't have a timestamp, and then sort by timestamps
    entries.retain(|entry| entry.published.is_some() || entry.updated.is_some());
    entries
        .sort_by_key(|entry| Reverse(entry.published.unwrap_or(entry.updated.unwrap_or_default())));

    let html_string = generate_html(entries).into_string();

    let output_path = Path::new("output/index.html");
    DirBuilder::new()
        .recursive(true)
        .create(output_path.parent().unwrap())
        .unwrap();

    match write(output_path, html_string) {
        Ok(_) => println!("Successfully wrote to {}", output_path.display()),
        Err(e) => eprintln!("Failed to write to {}: {}", output_path.display(), e),
    }

    Ok(())
}
